#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# CREATED: 10/11/16 6:07 PM by Justin Salamon <justin.salamon@nyu.edu>

'''Exception classes for Scaper'''


class ScaperError(Exception):
    '''The root Scaper exception class'''
    pass
